/*
 * acpidmp.c - ACPI table dumper
 *
 * Copyright (C) 1999 Andrew Henroid
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <linux/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define ACPI_BIOS_ROM_BASE 0x000e0000
#define ACPI_BIOS_ROM_END  0x00100000
#define ACPI_BIOS_ROM_SIZE (ACPI_BIOS_ROM_END - ACPI_BIOS_ROM_BASE)

#define ACPI_RSDP1_SIG 0x20445352 /* 'RSD ' */
#define ACPI_RSDP2_SIG 0x20525450 /* 'PTR ' */
#define ACPI_FACP_SIG  0x50434146 /* 'FACP' */

#define PAGE_OFFSET(addr) ((unsigned long)(addr) & (getpagesize() - 1))

struct acpi_rsdp
{
    __u32 signature[2];
    __u8 checksum;
    __u8 oem[6];
    __u8 reserved;
    __u32 rsdt;
};

struct acpi_table
{
    __u32 signature;
    __u32 length;
    __u8 rev;
    __u8 checksum;
    __u8 oem[6];
    __u8 oem_table[8];
    __u32 oem_rev;
    __u32 creator;
    __u32 creator_rev;
};

struct acpi_facp
{
    struct acpi_table hdr;
    __u32 facs;
    __u32 dsdt;
    __u8 int_model;
    __u8 reserved;
    __u16 sci_int;
    __u32 smi_cmd;
    __u8 acpi_enable;
    __u8 acpi_disable;
    __u8 s4bios_req;
    __u8 reserved2;
    __u32 pm1a_evt;
    __u32 pm1b_evt;
    __u32 pm1a_cnt;
    __u32 pm1b_cnt;
    __u32 pm2_cnt;
    __u32 pm_tmr;
    __u32 gpe0;
    __u32 gpe1;
    __u8 pm1_evt_len;
    __u8 pm1_cnt_len;
    __u8 pm2_cnt_len;
    __u8 pm_tm_len;
    __u8 gpe0_len;
    __u8 gpe1_len;
    __u8 gpe1_base;
    __u8 reserved3;
    __u16 p_lvl2_lat;
    __u16 p_lvl3_lat;
    __u16 flush_size;
    __u16 flush_stride;
    __u8 duty_offset;
    __u8 duty_width;
    __u8 day_alarm;
    __u8 mon_alarm;
    __u8 century;
    __u8 reserved4;
    __u8 reserved5;
    __u8 reserved6;
    __u32 flags;
};

/*
 * Map an ACPI table into virtual memory
 */
static struct acpi_table*
acpi_map_table(int mem, unsigned long addr)
{
    /* mmap header to determine table size */
    struct acpi_table *table = NULL;
    unsigned long offset = PAGE_OFFSET(addr);
    __u8 *mapped = mmap(NULL,
                        sizeof(struct acpi_table) + offset,
                        PROT_READ,
                        MAP_PRIVATE,
                        mem,
                        (unsigned long) addr - offset);
    table = (struct acpi_table*)
        ((mapped != MAP_FAILED) ? (mapped + offset):NULL);
    if(table)
    {
        /* re-mmap entire table */
        unsigned long size = table->length;
        munmap((__u8*) table - offset, sizeof(struct acpi_table) + offset);
        mapped = mmap(NULL,
                      size + offset,
                      PROT_READ,
                      MAP_PRIVATE,
                      mem,
                      (unsigned long) addr - offset);
        table = (struct acpi_table*)
            ((mapped != MAP_FAILED) ? (mapped + offset):NULL);
    }
    return table;
}

/*
 * Unmap an ACPI table from virtual memory
 */
static void
acpi_unmap_table(struct acpi_table *table)
{
    if(table)
    {
        unsigned long offset = PAGE_OFFSET(table);
        munmap((__u8*) table - offset, table->length + offset);
    }
}

/*
 * Locate the RSDP
 */
static unsigned long
acpi_find_rsdp(int mem,
               unsigned long base,
               unsigned long size,
               struct acpi_rsdp *rsdp)
{
    unsigned long addr = 0;
    __u8 *mapped = mmap(NULL, size, PROT_READ, MAP_PRIVATE, mem, base);
    if(mapped != MAP_FAILED)
    {
        __u8 *i;
        for(i = mapped; i < mapped + size; i += 16)
        {
            struct acpi_rsdp *r = (struct acpi_rsdp*) i;
            if(r->signature[0] == ACPI_RSDP1_SIG
               && r->signature[1] == ACPI_RSDP2_SIG)
            {
                memcpy(rsdp, r, sizeof(*rsdp));
                addr = base + (i - mapped);
                break;
            }
        }
        munmap(mapped, size);
    }
    return addr;
}

/*
 * Output data
 */
static void
acpi_show_data(FILE *fp, __u8 *data, int size)
{
    int i, remain = size;

    while(remain > 0)
    {
        fprintf(fp, "  %04x:", size - remain);
        for(i = 0; i < 16 && i < remain; i++)
        {
            fprintf(fp, " %02x", data[i]);
        }
        for(; i < 16; i++)
        {
            fprintf(fp, "   ");
        }
        fprintf(fp, "  ");
        for(i = 0; i < 16 && i < remain; i++)
        {
            fputc((data[i] >= ' ' && data[i] <= '~') ? data[i]:'.', fp);
        }
        fprintf(fp, "\n");
        data += 16;
        remain -= 16;
    }
}

/*
 * Output ACPI table
 */
static void
acpi_show_table(FILE *fp,
                struct acpi_table* table,
                unsigned long addr,
                __u32 search)
{
    if(!search)
    {
        char sig[5];
        memcpy(sig, &table->signature, 4);
        sig[4] = '\0';
        fprintf(fp, "%s @ 0x%08lx\n", sig, addr);
        acpi_show_data(fp, (__u8*) table, table->length);
        fprintf(fp, "\n");
    }
    else if(table->signature == search)
    {
        fwrite(table, 1, table->length, fp);
    }
}

int
main(int argc, char **argv)
{
    int mem, i;
    __u32 search;
    unsigned long facp_addr;
    unsigned long dsdt_addr;
    struct acpi_rsdp rsdp;
    unsigned long rsdp_addr;
    struct acpi_table *rsdt;
    int rsdt_entry_count;
    __u32 *rsdt_entry;
    char oem[7];

    search = 0;
    if(argv[1] && strlen(argv[1]) == 4)
    {
        memcpy(&search, argv[1], sizeof(search));
    }

    mem = open("/dev/mem", O_RDONLY);
    if(mem < 0)
    {
        fprintf(stderr, "acpidmp: cannot open /dev/mem (Running as root?)\n");
        exit(1);
    }

    rsdp_addr = acpi_find_rsdp(mem,
                               ACPI_BIOS_ROM_BASE,
                               ACPI_BIOS_ROM_SIZE,
                               &rsdp);
    if(!rsdp_addr)
    {
        fprintf(stderr, "acpidmp: cannot find an RSDP (Is ACPI enabled?)\n");
        exit(1);
    }

    memcpy(oem, rsdp.oem, 6);
    oem[6] = '\0';
    for(i = 5; i > 0 && (oem[i] == '\0' || oem[i] == ' '); i--)
    {
        oem[i] = '\0';
    }
    if(!search)
    {
        printf("RSDP \"%s\" @ 0x%08lx\n", oem, rsdp_addr);
        acpi_show_data(stdout, (__u8*) &rsdp, sizeof(rsdp));
        printf("\n");
    }
    
    rsdt = acpi_map_table(mem, rsdp.rsdt);
    if(!rsdt)
    {
        fprintf(stderr, "acpidmp: cannot map the RSDT\n");
        exit(1);
    }
    acpi_show_table(stdout, rsdt, rsdp.rsdt, search);

    facp_addr = 0;
    dsdt_addr = 0;

    rsdt_entry = (__u32*)(rsdt + 1);
    rsdt_entry_count = (rsdt->length - sizeof(*rsdt)) >> 2;
    while(rsdt_entry_count)
    {
        struct acpi_table *dt = acpi_map_table(mem, *rsdt_entry);
        if(dt)
        {
            acpi_show_table(stdout, dt, *rsdt_entry, search);
            if(dt->signature == ACPI_FACP_SIG)
            {
                struct acpi_facp *facp;
                struct acpi_table *dsdt, *facs;
                
                facp = (struct acpi_facp*) dt;
                facp_addr = *rsdt_entry;

                dsdt = acpi_map_table(mem, facp->dsdt);
                if(dsdt)
                {
                    dsdt_addr = facp->dsdt;
                    acpi_show_table(stdout, dsdt, facp->dsdt, search);
                    acpi_unmap_table(dsdt);
                }

                facs = acpi_map_table(mem, facp->facs);
                if(facs)
                {
                    acpi_show_table(stdout, facs, facp->facs, search);
                    acpi_unmap_table(facs);
                }
            }
            
            acpi_unmap_table(dt);
        }
        rsdt_entry++;
        rsdt_entry_count--;
    }
    acpi_unmap_table(rsdt);

    close(mem);

    if(!facp_addr)
    {
        fprintf(stderr, "acpidmp: cannot find an FACP\n");
        exit(1);
    }
    if(!dsdt_addr)
    {
        fprintf(stderr, "acpidmp: cannot find a DSDT\n");
        exit(1);
    }

    return 0;
}
